/* This file is part of the Palabos library.
 * Copyright (C) 2009, 2010 Jonas Latt
 * E-mail contact: jonas@lbmethod.org
 * The most recent release of Palabos can be downloaded at
 * <http://www.lbmethod.org/palabos/>
 *
 * The library Palabos is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * The library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/** \file
 * Helper functions for domain initialization -- header file.
 */
#ifndef THROMBOSIS_FUNCTIONAL_3D_HH
#define THROMBOSIS_FUNCTIONAL_3D_HH

#include "thrombosisFunctional3D.h"
#include "core/plbDebug.h"
#include "core/util.h"
#include "core/blockStatistics.h"
#include "latticeBoltzmann/momentTemplates.h"
#include "latticeBoltzmann/geometricOperationTemplates.h"
#include "finiteDifference/fdStencils1D.h"
#include "atomicBlock/atomicBlock3D.h"
#include "atomicBlock/blockLattice3D.h"
#include "atomicBlock/dataField3D.h"
//#include "thrombosisLattices3D.h"
#include "nodeNatureTemplates.h"
#include <cmath>
#include <limits>

namespace plb {

/* ******** CopyVelocityToParticlesFunctional3D ****************************************** */

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::
    CopyVelocityToParticlesFunctional3D(Array<T1,3> &settlingVelocityVector_) :
        settlingVelocityVector(settlingVelocityVector_)
{  }


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::process(
            Box3D domain,
            BlockLattice3D<T1,Descriptor1> &lattice,
            BlockLattice3D<T2,Descriptor2> &particles)
{
    enum { vel = Descriptor2<T2>::ExternalField::velocityBeginsAt};

    BounceBack<T1,Descriptor1> BBdynamics;
    NoDynamics<T1,Descriptor1> NNdynamics;

    //pcout << "settlingVelocityVector[2] = " << settlingVelocityVector[2] << std::endl;

    Dot3D off = computeRelativeDisplacement(lattice, particles);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX)
    {
        plint oX = iX + off.x; //oX, oY and oZ are used below to assign a new velocity to particles (u.to_cArray)
        for (plint iY=domain.y0; iY<=domain.y1; ++iY)
        {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ)
            {
                plint oZ = iZ + off.z;
                const Cell<T1,Descriptor1> &cell = lattice.get(iX,iY,iZ);

                Array<T1,Descriptor1<T1>::d> u;

                if ( cell.getDynamics().getId() == BBdynamics.getId() || cell.getDynamics().getId() == NNdynamics.getId() )
                {
                    u[0] = T1();
                    u[1] = T1();
                    u[2] = T1();

                }
                else
                {
                    cell.computeVelocity(u);

                    u[0] = u[0] + settlingVelocityVector[0];
                    u[1] = u[1] + settlingVelocityVector[1];
                    u[2] = u[2] + settlingVelocityVector[2];
                }


                u.to_cArray(particles.get(oX,oY,oZ).getExternal(vel));

            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>*
        CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::clone() const
{
    return new CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void CopyVelocityToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
}

/* ******** MakeThrombosisToParticlesFunctional3D ****************************************** */

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::
    MakeThrombosisToParticlesFunctional3D(T1 omegaAD_, T1 iniParticles_, Dynamics<T2,Descriptor2>* dynamicsParticles_) :
        omegaAD(omegaAD_), iniParticles(iniParticles_), dynamicsParticles(dynamicsParticles_)
{  }


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::process(
    Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice,
    BlockLattice3D<T2,Descriptor2> &particles)
{
    enum { nodeNature = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           omegaNS = Descriptor1<T1>::ExternalField::omegaBeginsAt};

    BounceBack<T2,Descriptor2> BBdynamics;

    Dot3D off = computeRelativeDisplacement(lattice,particles);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + off.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + off.z;

                const Cell<T1,Descriptor1> &cell = lattice.get(iX,iY,iZ);
		        // in this function I could probably also update the state of justBecameFluid nodes.

                // If I'm bounce back and the cell next to me is either fluid ...
                // then I'm a solid node next to the bulk. or nextToWall lattice.

                if ( *(cell.getExternal(nodeNature)) == NodeNature::notErodableWall ||
                     *(cell.getExternal(nodeNature)) == NodeNature::erodableWall )
                {
                    bool myNeighBorIsNextToWall = false;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop)
                    {
                        plint nextX = iX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        if (*(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::nextToWall ||
                            *(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::bulk)
                        {
                            myNeighBorIsNextToWall = true;
                            break;
                        }
                    }

                    if (myNeighBorIsNextToWall)
                        particles.attributeDynamics(oX,oY,oZ, dynamicsParticles->clone());
                    else
                        particles.attributeDynamics(oX,oY,oZ, new NoDynamics<T2,Descriptor2>());



		        }

            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>*
        MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::clone() const
{
    return new MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void MakeThrombosisToParticlesFunctional3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
}

/* ******** MakeThrombosisFromWssFunctional3D ****************************************** */

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
    MakeThrombosisFromWssFunctional3D(T1 stressThreshold_, T1 criticalTime_, T1 criticalNumber_, T1 prob_) :
        stressThreshold(stressThreshold_), criticalTime(criticalTime_), criticalNumber(criticalNumber_), prob(prob_)
{  }

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
        processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks)
{
    enum { nodeNature = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor1<T1>::ExternalField::timeBeginsAt};

    using namespace freeSurfaceFlag;


    BlockLattice3D<T1,Descriptor1>& lattice =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    BlockLattice3D<T2,Descriptor2>& particles =
            *dynamic_cast<BlockLattice3D<T2,Descriptor2> *>(atomicBlocks[1]);

    ScalarField3D<T3>& temporalMeanWss =
            *dynamic_cast<ScalarField3D<T3> *>(atomicBlocks[2]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[3]);

    ScalarField3D<T1>& mass =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[4]);

    ScalarField3D<T1>& volumeFraction =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[5]);

     ScalarField3D<T1>& justDeposited =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[6]);

    Dot3D offset = computeRelativeDisplacement(lattice, particles);
    Dot3D offWss = computeRelativeDisplacement(lattice, temporalMeanWss);
    Dot3D offmass = computeRelativeDisplacement(lattice, mass);
    Dot3D offvolFrac = computeRelativeDisplacement(lattice, volumeFraction);
    Dot3D offsetFlag = computeRelativeDisplacement(lattice, flag);
    Dot3D offsetJustDeposited = computeRelativeDisplacement(lattice, justDeposited);

    for (plint iX=domain.x0; iX<=domain.x1; ++iX)
    {
        plint oX = iX + offset.x;
        plint wX = iX + offWss.x;
        plint massX = iX + offmass.x;
        plint volFracX = iX + offvolFrac.x;
	    plint flagX = iX + offsetFlag.x;
        plint jDepX = iX + offsetJustDeposited.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY)
        {
            plint oY = iY + offset.y;
            plint wY = iY + offWss.y;
            plint massY = iY + offmass.y;
            plint volFracY = iY + offvolFrac.y;
	        plint flagY = iY + offsetFlag.y;
            plint jDepY = iY + offsetJustDeposited.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ)
            {
                plint oZ = iZ + offset.z;
                plint volFracZ = iZ + offvolFrac.z;
                plint wZ = iZ + offWss.z;
                plint massZ = iZ + offmass.z;
		        plint flagZ = iZ + offsetFlag.z;
		        plint jDepZ = iZ + offsetJustDeposited.z;

                if (*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::nextToWall &&
		            flag.get(flagX,flagY,flagZ) == fluid)
		        {
		            //std::cout << "Do I enter here? MakeThrombosisFromWssFunctional3D" << std::endl;

                    T1 numberParticles = particles.get(oX,oY,oZ).computeDensity();
                    Array<T1,3> averageVelocity ( T1(0), T1(0), T1(0));
                    plint counterOfCoolneighs = 0;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop)
                    {

                        plint nextX = iX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        plint nextX_f = flagX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY_f = flagY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ_f = flagZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        if (*(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::nextToWall &&
                            flag.get(nextX_f,nextY_f,nextZ_f) == fluid )
                        {

                            counterOfCoolneighs +=1;

                            Array<T2,3> localVelocity;

                            lattice.get(nextX,nextY,nextZ).computeVelocity(localVelocity);
                            averageVelocity[0] +=  localVelocity[0];
                            averageVelocity[1] +=  localVelocity[1];
                            averageVelocity[2] +=  localVelocity[2];

                        }
                        if (counterOfCoolneighs >0)
                        {
                            averageVelocity[0] /=  counterOfCoolneighs;
                            averageVelocity[1] /=  counterOfCoolneighs;
                            averageVelocity[2] /=  counterOfCoolneighs;
                        }
                        else
                        {
                            averageVelocity[0] =   T1(0);
                            averageVelocity[1] =   T1(0);
                            averageVelocity[2] =   T1(0);
                        }

		            }


                    if ( temporalMeanWss.get(wX,wY,wZ) <= stressThreshold ) {
                        *(lattice.get(iX,iY,iZ).getExternal(time)) += 1;
                    }
                    else{
			            *(lattice.get(iX,iY,iZ).getExternal(time)) = 0;
		                }
                    // if the shear stress is low enough for a sufficiently long time and if there are enough particles
                    if (*(lattice.get(iX,iY,iZ).getExternal(time)) >= criticalTime && numberParticles >= criticalNumber) {
                        // ... you draw a random number
                        // T2 rn = std::rand() / (T2)RAND_MAX;
//                         if (rn >= prob) {
                            // and thrombose is done with a given probability
                            lattice.attributeDynamics(iX,iY,iZ, new BounceBack<T1,Descriptor1>((T1)lattice.getTimeCounter().getTime()));
                            *(lattice.get(iX,iY,iZ).getExternal(time)) = 0;
                            //std::cout << "we have some newDeposit" << std::endl;
                            // change the flag
                            flag.get(flagX,flagY,flagZ) = (int)freeSurfaceFlag::wall;
                            mass.get(massX,massY,massZ) = T1(0);
                            volumeFraction.get(volFracX,volFracY,volFracZ) = T1(0);

                            *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;

                            justDeposited.get(jDepX,jDepY,jDepZ) = T1(-50);

                            // particle cells have already the right dynamics.



                            // I could change the particles dynamic here and then fix is density (number of particles)

                             //std::cout << "we have some new cell belonging to the deposit" << std::endl ;
//                             *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::wall;
                        //}
                    }
                }
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>*
    MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::clone() const
{
    return new MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void MakeThrombosisFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dataStructure;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::nothing;
    modified[2] = modif::nothing;
    modified[3] = modif::staticVariables;
    modified[4] = modif::staticVariables;
    modified[5] = modif::staticVariables;
    modified[6] = modif::staticVariables;
}


/* ******** SetParticleNumber3D ****************************************** */

template<typename T, template<class U> class Descriptor>
void SetParticleNumber3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice)
{
    using namespace descriptors;
//     enum {
//         nodeNature = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
//     };

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
//                 if (lattice.get(iX,iY,iZ).getExternal(nodeNature) == NodeNature::bulk) {
                for (plint iPop = 1; iPop < Descriptor<T>::q; ++iPop)
		    lattice.get(iX,iY,iZ)[iPop] = Descriptor<T>::rhoBar(T());
                lattice.get(iX,iY,iZ)[0] = Descriptor<T>::rhoBar(number);
//                 }
            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
SetParticleNumber3D<T,Descriptor>* SetParticleNumber3D<T,Descriptor>::clone() const {
    return new SetParticleNumber3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void SetParticleNumber3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
}

/* ******** ZeroGradientParticleNumber3D ****************************************** */

template<typename T, template<class U> class Descriptor, int direction, int orientation>
void ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice)
{
    using namespace descriptors;
//     enum {
//         nodeNature = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
//     };

    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
//                 if (lattice.get(iX,iY,iZ).getExternal(nodeNature) == NodeNature::bulk) {
                T rho = lattice.get (
                    iX+(direction==0 ? (-orientation):0),
                    iY+(direction==1 ? (-orientation):0),
                    iZ+(direction==2 ? (-orientation):0)  ).computeDensity();

                for (plint iPop = 1; iPop < Descriptor<T>::q; ++iPop) lattice.get(iX,iY,iZ)[iPop] = Descriptor<T>::rhoBar(T());
                lattice.get(iX,iY,iZ)[0] = Descriptor<T>::rhoBar(rho);
//                 }
            }
        }
    }
}

template<typename T, template<class U> class Descriptor, int direction, int orientation>
ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>* ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>::clone() const {
    return new ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>(*this);
}

template<typename T, template<class U> class Descriptor, int direction, int orientation>
void ZeroGradientParticleNumber3D<T,Descriptor,direction,orientation>::
    getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::dataStructure;
}

///////////////  Erosion Process   ////////////////////////////////////

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
    ErosionFromWssFunctional3D(T1 stressThreshold_, T1 criticalTime_, T1 criticalNumber_, T1 prob_, T1 omegaNS_, T1 cSmago_,
			       Array<T1,3> gravity_, Dynamics<T1,Descriptor1>* dynamics_, Dynamics<T2,Descriptor2>* dynamicsParticles_) :
        stressThreshold(stressThreshold_), criticalTime(criticalTime_), criticalNumber(criticalNumber_), prob(prob_),
        omegaNS(omegaNS_), cSmago(cSmago_),
        gravity(gravity_), dynamics(dynamics_), dynamicsParticles(dynamicsParticles_)
{  }

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::
        processGenericBlocks(Box3D domain, std::vector<AtomicBlock3D*> atomicBlocks)
{
    enum { nodeNature  = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           timeErosion = Descriptor1<T1>::ExternalField::timeErosionBeginsAt,
           externalForce = Descriptor1<T1>::ExternalField::forceBeginsAt,
    };

    using namespace freeSurfaceFlag;


    BlockLattice3D<T1,Descriptor1>& lattice =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    BlockLattice3D<T2,Descriptor2>& particles =
            *dynamic_cast<BlockLattice3D<T2,Descriptor2> *>(atomicBlocks[1]);

    ScalarField3D<T3>& temporalMeanWss =
            *dynamic_cast<ScalarField3D<T3> *>(atomicBlocks[2]);


    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[3]);

    ScalarField3D<T1>& mass =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[4]);

    ScalarField3D<T1>& volumeFraction =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[5]);

    ScalarField3D<T1>& justEroded =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[6]);


    Dot3D offset = computeRelativeDisplacement(lattice, particles);
    Dot3D offWss = computeRelativeDisplacement(lattice, temporalMeanWss);
    Dot3D offmass = computeRelativeDisplacement(lattice, mass);
    Dot3D offvolFrac = computeRelativeDisplacement(lattice, volumeFraction);
    Dot3D offsetFlag = computeRelativeDisplacement(lattice, flag);
    Dot3D offsetJustEroded = computeRelativeDisplacement(lattice, justEroded);



    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + offset.x;
        plint wX = iX + offWss.x;
        plint massX = iX + offmass.x;
	    plint flagX = iX + offsetFlag.x;
        plint volFracX = iX + offvolFrac.x;
        plint jEroX = iX + offsetJustEroded.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY            = iY + offset.y;
            plint wY            = iY + offWss.y;
            plint massY         = iY + offmass.y;
	        plint flagY         = iY + offsetFlag.y;
            plint volFracY      = iY + offvolFrac.y;
            plint jEroY         = iY + offsetJustEroded.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + offset.z;
                plint wZ = iZ + offWss.z;
                plint massZ = iZ + offmass.z;
		        plint flagZ = iZ + offsetFlag.z;
		        plint volFracZ = iZ + offvolFrac.z;
                plint jEroZ = iZ + offsetJustEroded.z;

                if (*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::erodableWall /*&&
                flag.get(flagX,flagY,flagZ) == wall*/ )
                {
                    //std::cout << "Do I enter here? ErosionFromWssFunctional3D" << std::endl;

                    // now we should get the wss acting on this node
                    // better way, probably, is to make an average over its neighs (easier in terms of position recon.)

                    Array<T1,3> averageVelocity ( T1(0), T1(0), T1(0));
                    plint counterOfCoolneighs = 0;
                    for (plint iPop = 1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop)
                    {

                        plint nextX = iX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY = iY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ = iZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        plint nextX_f = flagX + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY_f = flagY + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ_f = flagZ + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        if (*(lattice.get(nextX,nextY,nextZ).getExternal(nodeNature)) == NodeNature::nextToWall &&
                            flag.get(nextX_f,nextY_f,nextZ_f) == fluid  )
                        {


                            counterOfCoolneighs +=1;

                            Array<T2,3> localVelocity;

                            lattice.get(nextX,nextY,nextZ).computeVelocity(localVelocity);
                            averageVelocity[0] +=  localVelocity[0];
                            averageVelocity[1] +=  localVelocity[1];
                            averageVelocity[2] +=  localVelocity[2];


                        }

                    }

                    if (counterOfCoolneighs > 0)
                    {
                        averageVelocity[0] /=  counterOfCoolneighs;
                        averageVelocity[1] /=  counterOfCoolneighs;
                        averageVelocity[2] /=  counterOfCoolneighs;
                    }
                    else
                    {
                        averageVelocity[0] =   T1(0);
                        averageVelocity[1] =   T1(0);
                        averageVelocity[2] =   T1(0);
                    }

                    if (temporalMeanWss.get(wX,wY,wZ) > stressThreshold)
                    {
                    //pcout << " if stress > threshold --> temporalMeanWss.get(wX,wY,wZ)= " << temporalMeanWss.get(wX,wY,wZ) << std::endl;
                        *(lattice.get(iX,iY,iZ).getExternal(timeErosion)) += 1;
                    }
                    else
                    {
                    //pcout << " if stress < threshold --> temporalMeanWss.get(wX,wY,wZ)= " << temporalMeanWss.get(wX,wY,wZ) << std::endl;
                        *(lattice.get(iX,iY,iZ).getExternal(timeErosion)) = 0;
                    }
                    // if the shear stress is low enough for a sufficiently long time and if there are enough particles
                    if (*(lattice.get(iX,iY,iZ).getExternal(timeErosion)) >= int (criticalTime) )
                    {
                        // ... you draw a random number
                        //     T2 rn = std::rand() / (T2)RAND_MAX;
                        //     if (rn >= prob) {
                        // and thrombose is done with a given probability
                        lattice.attributeDynamics(iX,iY,iZ, new SmagorinskyBGKdynamics<T1,Descriptor1>(omegaNS, cSmago));
                        //lattice.attributeDynamics(iX,iY,iZ, dynamics->clone() );
                        //std::cout << "well and here we erode a new cell" << std::endl;

                        // what about the other fields for a free surface simulation? are they update in some other dataProcessor of the
                        // freeSurface Scheme?
                        iniCellAtEquilibrium(lattice.get(iX,iY,iZ), T1(1), averageVelocity);
                        //*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::nextToWall; // or even bulk?

                        T1 *force = lattice.get(iX,iY,iZ).getExternal(externalForce);
                        force[0] = gravity[0];
                        force[1] = gravity[1];
                        force[2] = gravity[2];
                        // change the flag
                        flag.get(flagX,flagY,flagZ) = fluid;
                        mass.get(massX,massY,massZ) = T1(1);

                        volumeFraction.get(volFracX,volFracY,volFracZ) = T1(1) ;
                        *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::bulk;
                        justEroded.get(jEroX,jEroY,jEroZ) = T1(50);
                        *(lattice.get(iX,iY,iZ).getExternal(timeErosion)) = 0;


                        particles.attributeDynamics(oX,oY,oZ, dynamicsParticles->clone());
                        iniCellAtEquilibrium(particles.get(oX,oY,oZ), criticalNumber, averageVelocity );

                            //}
                    }
                }
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>*
    ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::clone() const
{
    return new ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2, typename T3>
void ErosionFromWssFunctional3D<T1,Descriptor1,T2,Descriptor2,T3>::getTypeOfModification (
        std::vector<modif::ModifT>& modified ) const
{
    modified[0] = modif::dataStructure;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
    modified[2] = modif::nothing;
    modified[3] = modif::staticVariables;
    modified[4] = modif::staticVariables;
    modified[5] = modif::staticVariables;
    modified[6] = modif::staticVariables;
}





///////////////////// ******************************************* ///////////////////////////////
/* ******** ChangeDynamicForAdvectionDiffusionDynamic3D ****************************************** */

/////////////////  THIS FUNCTION IS ONLY FOR INTERFACE DYNAMICS ..
template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>::
        ChangeDynamicForAdvectionDiffusionDynamic3D(T2 omegaAD_, Dynamics<T2,Descriptor2>* dynamics_):
	omegaAD(omegaAD_), dynamics(dynamics_)
{ }


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>::processGenericBlocks (
        Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks )
{
    PLB_ASSERT( atomicBlocks.size()==3 );
    global::timer("ChangeDynamicForAdvectionDiffusionDynamic3D").start();
    // typedef Descriptor1<T1> D1;
    // typedef Descriptor2<T2> D2;
    using namespace freeSurfaceFlag;

    enum { nodeNature  		= Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           timeErosion 		= Descriptor1<T1>::ExternalField::timeErosionBeginsAt,
           externalForce 	= Descriptor1<T1>::ExternalField::forceBeginsAt,
           omegaSmago  		= Descriptor1<T1>::ExternalField::omegaBeginsAt
    };

    // we need the fluid lattice (both for dynamic and update after collide and Stream)
    // the flag scalarField (in order to know the interface position)
    // and the particles (advection diffusion scheme)


    BlockLattice3D<T1,Descriptor1>& lattice =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    BlockLattice3D<T2,Descriptor2>& particles =
            *dynamic_cast<BlockLattice3D<T2,Descriptor2> *>(atomicBlocks[1]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[2]);


    Dot3D offLF  = computeRelativeDisplacement(lattice,flag);    // lattice-flag

    Dot3D offLP = computeRelativeDisplacement(lattice,particles);    // lattice-advection diffusion



    for (plint iX=domain.x0; iX<=domain.x1; ++iX)
    {
        plint oX_LP  = iX + offLP.x;
	    plint oX_LF  = iX + offLF.x;
	    for (plint iY=domain.y0; iY<=domain.y1; ++iY)
        {
            plint oY_LP  = iY + offLP.y;
	        plint oY_LF  = iY + offLF.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ)
            {
		        plint oZ_LP  = iZ + offLP.z;
		        plint oZ_LF  = iZ + offLF.z;

        		//const Cell<T1,Descriptor1> &cell = lattice.get(iX,iY,iZ);
		        int flagSi = flag.get(oX_LF,oY_LF,oZ_LF);
                if (flagSi == fluid || flagSi == interface )
                {
		            particles.attributeDynamics(oX_LP,oY_LP,oZ_LP, dynamics->clone());
                }
                else if (flagSi == empty)
                {
                    particles.attributeDynamics(oX_LP,oY_LP,oZ_LP, new NoDynamics<T2,Descriptor2>());
                }


            }
        }
    }

    global::timer("ChangeDynamicForAdvectionDiffusionDynamic3D").stop() ;

}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>*
        ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>::clone() const
{
    return new ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void ChangeDynamicForAdvectionDiffusionDynamic3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::dataStructure;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[2] = modif::nothing;
}
///////////////////// ******************************************* ///////////////////////////////



//////////////////  Boundary condition for Advection Diffusion. copy the unknow particle distribution functions //////////////


template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>::processGenericBlocks (
        Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks )
{
    PLB_ASSERT( atomicBlocks.size()==3 );
    global::timer("CopyUnknownFromNeighsFunctional3D").start();
    // typedef Descriptor1<T1> D1;
    // typedef Descriptor2<T2> D2;
    using namespace freeSurfaceFlag;


    BlockLattice3D<T1,Descriptor1>& particles =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    BlockLattice3D<T2,Descriptor2>& lattice =
            *dynamic_cast<BlockLattice3D<T2,Descriptor2> *>(atomicBlocks[1]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[2]);



    BounceBack<T2,Descriptor2> BBdynamics; // BounceBack dyamics per fluid
    NoDynamics<T1,Descriptor1> NoDynamicsParticles; // BounceBack dyamics per fluid

    Dot3D offPF  = computeRelativeDisplacement(particles,flag);    // lattice-flag
    Dot3D offPL = computeRelativeDisplacement(particles,lattice);    // lattice-advection diffusion

    for (plint iX=domain.x0; iX<=domain.x1; ++iX)
    {
        plint oX_PL  = iX + offPL.x;
	    plint oX_PF  = iX + offPF.x;

	    for (plint iY=domain.y0; iY<=domain.y1; ++iY)
        {
	        plint oY_PL  = iY + offPL.y;
	        plint oY_PF  = iY + offPF.y;

            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ)
            {
                plint oZ_PL  = iZ + offPL.z;
                plint oZ_PF  = iZ + offPF.z;

                // if the lattice node is bounce Back... we perform the boundaryCondition for the AdvectionDiffusion Scheme
                if ( ( lattice.get(oX_PL,oY_PL,oZ_PL).getDynamics().getId() == BBdynamics.getId() &&
                               particles.get(iX,iY,iZ).getDynamics().getId() != NoDynamicsParticles.getId() ) ||
                     flag.get(oX_PF,oY_PF,oZ_PF) == interface )
                {

                    // I'm not really sure about the physical meaning of the boundary condition I'm implementing here.
                    for(plint iPop=1; iPop > Descriptor1<T1>::q; ++iPop)
                    {
                        plint nextX = iX + Descriptor1<T1>::c[iPop][0];
                        plint nextY = iY + Descriptor1<T1>::c[iPop][1];
                        plint nextZ = iZ + Descriptor1<T1>::c[iPop][2];

                        plint nextX_f = oX_PF + Descriptor1<T1>::c[iPop][0];
                        plint nextY_f = oY_PF + Descriptor1<T1>::c[iPop][1];
                        plint nextZ_f = oZ_PF + Descriptor1<T1>::c[iPop][2];

                        if ( flag.get(nextX_f,nextY_f,nextZ_f) == fluid )
                        {
                            particles.get(iX,iY,iZ)[iPop] = particles.get(nextX,nextY,nextZ)[iPop];
                        }
                    }
		        }
            }
        }
    }

    global::timer("CopyUnknownFromNeighsFunctional3D").stop() ;

}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>*
        CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>::clone() const
{
    return new CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void CopyUnknownFromNeighsFunctional3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::dataStructure;
    modified[1] = modif::nothing;
    modified[2] = modif::nothing;
}
///////////////////// ******************************************* ///////////////////////////////

////////////////////////  ...



template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>::process(
    Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice,
    BlockLattice3D<T2,Descriptor2> &particles)
{
    enum { omegaNS = Descriptor1<T1>::ExternalField::omegaBeginsAt };

    Dot3D off = computeRelativeDisplacement(lattice,particles);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + off.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + off.z;
                const Cell<T1,Descriptor1> &cell = lattice.get(iX,iY,iZ);
		        // in this function I could probably also update the state of justBecameFluid nodes.
                if (lattice.get(iX,iY,iZ).getDynamics().getId() == meta::dynamicsRegistration<T2,Descriptor2>().getId("BGK_Smagorinsky"))
		        {
		            particles.attributeDynamics(oX,oY,oZ, new AdvectionDiffusionRLBdynamics<T2,Descriptor2>(*(cell.getExternal(omegaNS))));
		        }
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>*
        TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>::clone() const
{
    return new TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>(*this);
}

template<typename T1, template<class U> class Descriptor1, typename T2, template<class U> class Descriptor2>
void TestTurbulentAdvectionDiffusionFunctional3D<T1,Descriptor1,T2,Descriptor2>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
}





//////////////////   copy flagMatrix from externalField

template<typename T1, template<class U> class Descriptor1>
void CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>::process(
    Box3D domain, BlockLattice3D<T1,Descriptor1> &lattice,  ScalarField3D<T1> &copyTo)
{
    enum { nodeNature           = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           timeErosion          = Descriptor1<T1>::ExternalField::timeErosionBeginsAt,
           omegaSmago           = Descriptor1<T1>::ExternalField::omegaBeginsAt};

    Dot3D off = computeRelativeDisplacement(lattice,copyTo);
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + off.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + off.z;
                const Cell<T1,Descriptor1> &cell = lattice.get(iX,iY,iZ);
                // in this function I could probably also update the state of justBecameFluid nodes.

                //if (*cell.getExternal(nodeNature) == NodeNature::nextToWall)
                    copyTo.get(oX,oY,oZ) = T1 ( *cell.getExternal(nodeNature) ) ;
//                 else
//                     copyTo.get(oX,oY,oZ) = 0.;
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1>
CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>*
        CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>::clone() const
{
    return new CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>(*this);
}

template<typename T1, template<class U> class Descriptor1>
void CopyScalarFromExternalFieldToScalarField<T1,Descriptor1>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    // Full dynamics object must be recreated, because this data processor
    //   re-attributes a new dynamics and acts on the bulk only.
    modified[1] = modif::dataStructure;
}






template<typename T1, template<class U> class Descriptor1>
void CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>::process(
    Box3D domain, BlockLattice3D<T1,Descriptor1> &particles,  ScalarField3D<T1> &copyTo)
{

    Dot3D off = computeRelativeDisplacement(particles,copyTo);
    NoDynamics<T1,Descriptor1> NNdynamics;


    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + off.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + off.z;
                const Cell<T1,Descriptor1> &cell = particles.get(iX,iY,iZ);
                // in this function I could probably also update the state of justBecameFluid nodes.

                if ( cell.getDynamics().getId() == NNdynamics.getId() )
                    copyTo.get(oX,oY,oZ) = T1(1) ;
                else
                    copyTo.get(oX,oY,oZ) = T1() ;

            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1>
CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>*
        CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>::clone() const
{
    return new CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>(*this);
}

template<typename T1, template<class U> class Descriptor1>
void CopyStateOfLocalDynamicToScalarField<T1,Descriptor1>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::dataStructure;
}


//////////////////////////////////////////////////***

/*
template<typename T1, template<class U> class Descriptor1>
void BounceBackMatrix<T1,Descriptor1>::process(Box3D domain,
                                               BlockLattice3D<T1,Descriptor1> &particles,
                                               ScalarField3D<T1> &copyTo)
{

    Dot3D off = computeRelativeDisplacement(particles,copyTo);
    NoDynamics<T1,Descriptor1> NNdynamics;
    BounceBack<T1,Descriptor1> BBdynamics;
    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
        plint oX = iX + off.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
            plint oY = iY + off.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint oZ = iZ + off.z;
                const Cell<T1,Descriptor1> &cell = particles.get(iX,iY,iZ);
                if ( cell.getDynamics().getId() == NNdynamics.getId() || cell.getDynamics().getId() == BBdynamics.getId() )
                    copyTo.get(oX,oY,oZ) = T1(1) ;
                else
                    copyTo.get(oX,oY,oZ) = T1() ;
            }
        }
    }
}

template<typename T1, template<class U> class Descriptor1>
BounceBackMatrix<T1,Descriptor1>* BounceBackMatrix<T1,Descriptor1>::clone() const
{
    return new BounceBackMatrix<T1,Descriptor1>(*this);
}

template<typename T1, template<class U> class Descriptor1>
void BounceBackMatrix<T1,Descriptor1>::getTypeOfModification(std::vector<modif::ModifT>& modified) const
{
    modified[0] = modif::nothing;
    modified[1] = modif::dataStructure;
}


*/
/////////////////////////////////******************














///////////////////////////



template<typename T1, template<class U> class Descriptor1>
void CalculateTemporalWssAverage3D<T1,Descriptor1>::processGenericBlocks (
        Box3D domain,std::vector<AtomicBlock3D*> atomicBlocks )
{
    PLB_ASSERT( atomicBlocks.size()==4 );
    global::timer("CalculateTemporalWssAverage3D").start();
    // typedef Descriptor1<T1> D1;
    using namespace freeSurfaceFlag;

    enum { nodeNature = Descriptor1<T1>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor1<T1>::ExternalField::timeBeginsAt
    };



    BlockLattice3D<T1,Descriptor1>& lattice =
            *dynamic_cast<BlockLattice3D<T1,Descriptor1> *>(atomicBlocks[0]);

    ScalarField3D<T1>& wss =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[1]);

    ScalarField3D<T1>& temporalMeanWss =
            *dynamic_cast<ScalarField3D<T1> *>(atomicBlocks[2]);

    ScalarField3D<int>& flag =
            *dynamic_cast<ScalarField3D<int> *>(atomicBlocks[3]);



    Dot3D off_wss  	= computeRelativeDisplacement(lattice,wss);    // lattice-flag
    Dot3D off_meanWss 	= computeRelativeDisplacement(lattice,temporalMeanWss);    // lattice-advection diffusion
    Dot3D off_flag	= computeRelativeDisplacement(lattice,flag);    // lattice-advection diffusion
    //Dot3D absoluteOffset = lattice.getLocation();
    for (plint iX=domain.x0; iX<=domain.x1; ++iX)
    {
        plint iX_wss  = iX + off_wss.x;
	    plint iX_meanWss  = iX + off_meanWss.x;
	    plint iX_f  = iX + off_flag.x ;
	    //plint absX = iX + absoluteOffset.x;

        for (plint iY=domain.y0; iY<=domain.y1; ++iY)
        {
            plint iY_wss  = iY + off_wss.y;
            plint iY_meanWss  = iY + off_meanWss.y;
            plint iY_f  = iY + off_flag.y ;
            //plint absY = iY + absoluteOffset.y;

            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ)
            {
                plint iZ_wss  = iZ + off_wss.z;
                plint iZ_meanWss  = iZ + off_meanWss.z;
                plint iZ_f  = iZ + off_flag.z ;
                //plint absZ = iZ + absoluteOffset.z;

                // I'm not really sure about the physical meaning of the boundary condition I'm implementing here.
                plint counter=0;
                T1 localWss = T1(0);


                if ( *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::erodableWall ||
                     *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::nextToWall ||
                     *(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::bulk)
                {

                    for(plint iPop=1; iPop < descriptors::D3Q27Descriptor<T1>::q; ++iPop) // ==> Descriptor
                    {
                        plint nextX_wss = iX_wss + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY_wss = iY_wss + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ_wss = iZ_wss + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        plint nextX_f = iX_f + descriptors::D3Q27Descriptor<T1>::c[iPop][0];
                        plint nextY_f = iY_f + descriptors::D3Q27Descriptor<T1>::c[iPop][1];
                        plint nextZ_f = iZ_f + descriptors::D3Q27Descriptor<T1>::c[iPop][2];

                        if ( flag.get(nextX_f,nextY_f,nextZ_f) == (int)freeSurfaceFlag::fluid )
                        {
                            localWss += wss.get(nextX_wss, nextY_wss, nextZ_wss);
                            counter++;
                        }
                        else
                        {
                            localWss += T1(0);
                        }
                    }

                    if (counter > 0)
                    {
                        temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) += localWss/counter;
                        //pcout << "counter = " << counter  << "; temporal Means = " << temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) << std::endl;
                    }
                    else
                    {
                        temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) += T1(0);
                        //pcout << "counter = " << counter  << "; temporal Means = " << temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) << std::endl;
                    }
                }
                else
                {
                    temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) = T1(0);
                }
                //pcout << "temporalMeans iX,iY,iZ: =" << absX << " " <<absY << " "<< absZ << " " << temporalMeanWss.get(iX_meanWss,iY_meanWss,iZ_meanWss) << std::endl;
            }
        }
    }

    global::timer("CalculateTemporalWssAverage3D").stop() ;

}

template<typename T1, template<class U> class Descriptor1>
CalculateTemporalWssAverage3D<T1,Descriptor1>*
        CalculateTemporalWssAverage3D<T1,Descriptor1>::clone() const
{
    return new CalculateTemporalWssAverage3D<T1,Descriptor1>(*this);
}

template<typename T1, template<class U> class Descriptor1>
void CalculateTemporalWssAverage3D<T1,Descriptor1>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::nothing;
    modified[1] = modif::nothing;
    modified[2] = modif::staticVariables;
    modified[3] = modif::nothing;
}
///////////////////// ******************************************* ///////////////////////////////











///////////////////



template<typename T, template<class U> class Descriptor>
ChangeErodableCellStatus3D<T,Descriptor>::
        ChangeErodableCellStatus3D(plint erodableWallLimit_):
        erodableWallLimit(erodableWallLimit_)
{ }


template<typename T, template<class U> class Descriptor>
void ChangeErodableCellStatus3D<T,Descriptor>::process(
    Box3D domain, BlockLattice3D<T,Descriptor>& lattice, ScalarField3D<int>& flag)
{
    using namespace freeSurfaceFlag;
    using namespace descriptors;
    enum { nodeNature = Descriptor<T>::ExternalField::nodeNatureBeginsAt,
           time = Descriptor<T>::ExternalField::timeBeginsAt
    };
    Dot3D offsetFlag = computeRelativeDisplacement(lattice, flag);



    for (plint iX=domain.x0; iX<=domain.x1; ++iX) {
	plint iXflag = iX + offsetFlag.x;
        for (plint iY=domain.y0; iY<=domain.y1; ++iY) {
	    plint iYflag = iY + offsetFlag.y;
            for (plint iZ=domain.z0; iZ<=domain.z1; ++iZ) {
                plint iZflag = iZ + offsetFlag.z;

                if (lattice.get(iX,iY,iZ).getDynamics().getId() ==
                    meta::dynamicsRegistration<T,Descriptor>().getId("BounceBack"))
                {
		    plint nextZ_flag = iZflag + 1;

		    plint counterNeighs = 0;
		    for (plint iPop=0; iPop < descriptors::D3Q19Descriptor<T>::q ; iPop++ )
		    {
			if (descriptors::D3Q19Descriptor<T>::c[iPop][2] == 0 )
			{
			    plint nextX_flag = iXflag + descriptors::D3Q19Descriptor<T>::c[iPop][0];
			    plint nextY_flag = iYflag + descriptors::D3Q19Descriptor<T>::c[iPop][1];

			    if (flag.get(nextX_flag,nextY_flag,nextZ_flag) == (int)freeSurfaceFlag::fluid )
				counterNeighs += 1;
			}
		    }

		    // should I do that by using the lattice dynamics?
		    if ( counterNeighs ==  9 )
			*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::erodableWall;
		    else
			*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) = NodeNature::notErodableWall;


		   /* if ( lattice.get(iX,iY,iZ+1).getDynamics().getId() ==
                            meta::dynamicsRegistration<T2,Descriptor2>().getId("BGK_Smagorinsky")  )
			*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::erodableWall;
		    else
			*(lattice.get(iX,iY,iZ).getExternal(nodeNature)) == NodeNature::notErodableWall;  */
		}

            }
        }
    }
}

template<typename T, template<class U> class Descriptor>
ChangeErodableCellStatus3D<T,Descriptor>* ChangeErodableCellStatus3D<T,Descriptor>::clone() const {
    return new ChangeErodableCellStatus3D<T,Descriptor>(*this);
}

template<typename T, template<class U> class Descriptor>
void ChangeErodableCellStatus3D<T,Descriptor>::getTypeOfModification(std::vector<modif::ModifT>& modified) const {
    modified[0] = modif::staticVariables;
    modified[1] = modif::nothing;
}











}  // namespace plb

#endif  // DATA_ANALYSIS_FUNCTIONAL_3D_HH
